/*
 * mySQLenum
 *  
 * Copyright (C) 2009 Andrea Fabrizi <andrea.fabrizi@gmail.com>
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *
 */
#include <stdio.h>
#include <sys/types.h> 
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <netdb.h>
#include <unistd.h>
#include <ctype.h>
#include "mysqlenum.h"
#include "base64.h"

/*
 * TCP Socket Connect
 *
 * Returns:
 *  OK: socket descriptor
 *  KO: 0
 */
int socket_connect(const char* hostname, unsigned short int port)
{
    struct sockaddr_in serv_addr;
    struct hostent *server;
    int sd;
    
    /* socket() */
    sd = socket(AF_INET, SOCK_STREAM, 0);
    if (sd < 0) 
        return 0;
    
    /* risolvo l'host */
    server = gethostbyname(hostname);
    if (server == NULL)
        return 0;
    
    bzero((char *) &serv_addr, sizeof(serv_addr));
    serv_addr.sin_family = AF_INET;
    bcopy((char *)server->h_addr, (char *)&serv_addr.sin_addr.s_addr, server->h_length);
    serv_addr.sin_port = htons(port);
    
    /* connect() */
    if (connect(sd,(struct sockaddr *)&serv_addr,sizeof(serv_addr)) < 0) 
        return 0;
   
    return sd;
}

/*
 * Closing socket...
 */
void socket_close(int sd)
{
    close(sd);
}


/*
 * Reads a stram from the socket, dinamically allocating the enough memory 
 * 
 * Returns:
 * OK: buffer pointer
 * KO: NULL
 *
 */
char *socket_read(int sd)
{
    char c[1],
         *buffer;
    size_t buffer_size,
           bcount = 0;
    
    buffer_size = 128;
    buffer = malloc(buffer_size);
    
    /* reads one byte at time */
    while (read(sd, c, 1))
    {
        buffer[bcount] = c[0];
        bcount++;
        
        /* if the allocated memory is full
         * will be extended...
         */
        if (bcount >= buffer_size - 1)
        {
            buffer_size = buffer_size + 128;
            buffer = realloc(buffer, buffer_size);
        }
    }
    
    buffer[bcount] = 0;
    
    return buffer;
}


/*
 * Encode the *spaces* in url format
 *
 */
char *urlencode(const char* url)
{
    int nspaces = 0,
        i,
        url_encoded_len,
        url_len;
    char *url_encoded,
         *xx_space = "%20";
    
    url_len = strlen(url);
    
    if (!url_len)
    {
        asprintf(&url_encoded, "");
        return url_encoded;
    }
    
    /* calcolo in numero di spazi presenti nella stringa
     * per poter poi allocare la memoria necessaria
     */
    for (i=0; i<url_len; i++)
        if (url[i] == ' ') nspaces++;
    
    url_encoded_len = url_len + nspaces*2 + 1;
    url_encoded = malloc(url_encoded_len);
    bzero(url_encoded, url_encoded_len);
    
    for (i=0; i<url_len; i++)
    {
        if (url[i] != ' ')
            strncat(url_encoded,url + i,1);
        else
            strncat(url_encoded, xx_space,3);
    }
    
    return url_encoded;
}


/*
 * Do an HTTP GET/POST request
 * 
 * Returns:
 * OK: pointer to server reponse
 * KO: NULL
 *
 */
char *http_request(const char* http_req, struct configuration *conf)
{
    int sd,
        coded_len;
    size_t request_size;
    char *request,
         *response,
         *cookie,
         *cookie_header,
         *basic_auth,
         *basic_auth_header,
         *url_request,
         *custom_header;
            
    /* connection */
    sd = socket_connect(conf->host, conf->port);
    if (!sd)
    {
        perror("\n *** Error connecting to host");
        return NULL;
    }
    
    /* url-encode url */
    url_request = urlencode(http_req);
    
    /* url-encode cookies */
    if (!str_empty(conf->httpcookie))
    {
        cookie = urlencode(conf->httpcookie);
        asprintf(&cookie_header, "Cookie: %s\r\n", cookie);
        free(cookie);
    }
    else cookie_header = "";

    /* custom header */
    if (!str_empty(conf->customheader))
        asprintf(&custom_header, "%s\r\n", conf->customheader);
    else custom_header = "";
    
    /* basic auth */
    if (!str_empty(conf->httpauth))
    {
        coded_len = Base64encode_len(strlen(conf->httpauth));
        basic_auth = malloc(coded_len + 1);
        bzero(basic_auth, coded_len + 1);
        Base64encode(basic_auth, conf->httpauth, strlen(conf->httpauth));
        asprintf(&basic_auth_header, "Authorization: Basic %s\r\n", basic_auth);
        free(basic_auth);
    }
    else basic_auth_header = "";

    /* building http GET request */
    if (conf->httpmethod == HTTP_GET)
    {   
        request_size = asprintf(&request, 
            "GET %s?%s HTTP/1.0\r\n"
            "Host: %s\r\n"
            "%s"
            "%s"
            "User-Agent: %s\r\n"
            "%s"
            "\r\n",
            conf->page,url_request,conf->host, cookie_header, basic_auth_header, conf->useragent,custom_header);   
    }
    
    /* building http POST request */
    else
    {       
        request_size = asprintf(&request, 
            "POST %s HTTP/1.0\r\n"
            "Host: %s\r\n"
            "%s"
            "%s"
            "User-Agent: %s\r\n"
            "%s"
            "Content-Type: application/x-www-form-urlencoded\r\n"
            "Content-Length: %d\r\n"
            "\r\n"
            "%s",
            conf->page, conf->host, cookie_header, basic_auth_header, conf->useragent, custom_header, strlen(url_request), url_request);   
    }
 
    debug_msg(VERBOSE_LEVEL_2, "HTTP REQUEST:\n%s\n", request);
    
    /* send request to server */
    if (write (sd, request, request_size) < 0)
    {
        perror("\n *** Error sending request to server");
        
        if (!str_empty(conf->httpcookie))
            free(cookie_header);
            
        if (!str_empty(conf->httpauth))
            free(basic_auth_header);

        if (!str_empty(conf->customheader))
            free(custom_header);
            
        return NULL;
    }
    
    /* read the response */
    response = socket_read(sd);
    
    debug_msg(VERBOSE_LEVEL_3, "HTTP RESPONSE:\n%s\n", response);

    requests_count++;
    bytes_received+= strlen(response);
    bytes_sent+=request_size;
    
    socket_close(sd);
    free(request);
    free(url_request);
    if (!str_empty(conf->httpcookie))
        free(cookie_header);
    if (!str_empty(conf->httpauth))
        free(basic_auth_header);
    if (!str_empty(conf->customheader))
        free(custom_header);
                
    return response;
}
